import zlib
import uuid
import json
import asyncio
from urllib import parse

import pcqq.client as cli
import pcqq.network as net
import pcqq.utils as utils
import pcqq.const as const
import pcqq.binary as binary


def text(text: str) -> bytes:
    writer = binary.Writer()

    data = text.encode()
    data_size = len(data)

    writer.write_byte(0x01)
    writer.write_int16(data_size + 3)
    writer.write_byte(0x01)
    writer.write_int16(data_size)
    writer.write(data)

    return writer.clear()


def face(face_id: int) -> bytes:
    writer = binary.Writer()

    writer.write_byte(0x02)
    writer.write_int16(1 + 3)
    writer.write_byte(0x01)
    writer.write_int16(1)
    writer.write_byte(face_id)

    return writer.clear()


async def at(user_id: int, group_id: int) -> bytes:
    name = "@" + await net.get_user_cache(
        user_id=user_id,
        group_id=group_id
    )

    writer = binary.Writer()
    writer.write_hex("00 01 00 00")
    writer.write_int16(len(name))
    writer.write_hex("00")
    writer.write_int32(user_id)
    writer.write_hex("00 00")
    data = writer.clear()
    name = name.encode()

    writer.__init__()
    writer.write_byte(0x01)
    writer.write_int16(len(name))
    writer.write(name)
    writer.write_byte(0x06)
    writer.write_int16(len(data))
    writer.write(data)
    data = writer.clear()

    writer.__init__()
    writer.write_byte(0x01)
    writer.write_int16(len(data))
    writer.write(data)
    return writer.clear() + text(" ")


def xml(xml_code: str) -> bytes:
    writer = binary.Writer()
    xml_code = xml_code.replace("&", "&amp;")
    xml_code = xml_code.replace("&#44;", ",")
    data = zlib.compress(xml_code.encode(), -1)

    writer.write_byte(0x14)
    writer.write_int16(len(data) + 11)
    writer.write_hex("01")
    writer.write_int16(len(data) + 1)
    writer.write_hex("01")
    writer.write(data)
    writer.write_hex("02 00 04 00 00 00 02")
    return writer.clear()


def music(
    title: str = "",
    content: str = "",
    url: str = "",
    audio: str = "",
    cover: str = ""
) -> bytes:
    xml_code = const.MUSIC_CODE.format(
        title=title,
        content=content,
        url=url,
        audio=audio,
        cover=cover
    )

    return xml(xml_code)

async def qqmusic(keyword: str):
    info = json.loads((await cli.webget(
        url="https://c.y.qq.com/soso/fcgi-bin/client_search_cp?w=" +
            parse.quote(keyword)
    )).content[9:-1])["data"]["song"]["list"][0]

    audio = (await cli.webget(
        url="https://u.y.qq.com/cgi-bin/musicu.fcg?data=" + parse.quote(
            json.dumps(
                {
                    "comm": {"uin": 0, "format": "json", "ct": 24, "cv": 0},
                    "req": {"module": "CDN.SrfCdnDispatchServer", "method": "GetCdnDispatch", "param": {"guid": "3982823384", "calltype": 0, "userip": ""}},
                    "req_0": {"module": "vkey.GetVkeyServer", "method": "CgiGetVkey", "param": {"guid": "3982823384", "songmid": [info["songmid"]], "songtype": [0], "uin": "0", "loginflag": 1, "platform": "20"}}
                }
            ))
    )).json()["req_0"]["data"]["midurlinfo"][0]["purl"]

    html = (await cli.webget(f"https://y.qq.com/n/ryqq/songDetail/{info['songmid']}")).content
    start = html.find(b'photo_new\\u002F')+15
    end = html.find(b"?max_age", start)

    return music(
        title=info["songname"],
        content=info["singer"][0]["name"],
        url=f"https://y.qq.com/n/yqq/song/{info['songmid']}.html",
        audio="http://dl.stream.qqmusic.qq.com/" + audio,
        cover="https://y.qq.com/music/photo_new/" + html[start:end].decode()
    )


class Image:
    def __init__(self, im_data: bytes) -> None:
        self.ukey = None
        self.picid = None
        self.finish = False

        self.data = im_data
        self.size = len(im_data)
        self.hash = utils.hashmd5(im_data)
        self.width, self.height = utils.img_size(im_data)
        self.uuid = "{%s}.jpg" % (str(uuid.UUID(bytes=self.hash)).upper())


async def image_group(group_id: int, image: Image) -> bytes:
    writer = binary.Writer()
    await net.upload_group_image(group_id, image)

    wait_times = 90
    while not image.finish:
        await asyncio.sleep(1)
        wait_times -= 1
        if not wait_times:
            raise Exception(f"获取图片 [PQ:image,file={image.hash.hex().upper()}] 超时")
    


    writer.write_hex("03 00 CB 02")
    writer.write_int16(len(image.uuid))
    writer.write(image.uuid.encode())
    writer.write_hex("04 00 04")

    writer.write_hex("84 74 B1 53 05 00 04 BC")
    writer.write_hex("EB 03 B7 06 00 04 00 00")
    writer.write_hex("00 50 07 00 01 43 08 00")
    writer.write_hex("00 09 00 01 01 0B 00 00")
    writer.write_hex("14 00 04 11 00 00 00 15")
    writer.write_hex("00 04 00 00 00 8B 16 00")
    writer.write_hex("04 00 00 00 81 18 00 04")
    writer.write_hex("00 00 0E D3 FF 00 5C 15")
    writer.write_hex("36 20 39 32 6B 41 31 43")
    writer.write_hex("38 34 37 34 62 31 35 33")
    writer.write_hex("62 63 65 62 30 33 62 37")
    writer.write_hex("20 20 20 20 20 20 35 30")
    writer.write_hex("20 20 20 20 20 20 20 20")
    writer.write_hex("20 20 20 20 20 20 20 20")

    writer.write(image.uuid.encode())
    writer.write_hex("41")
    return writer.clear()


async def image_friend(user_id: int, image: Image) -> bytes:
    writer = binary.Writer()
    await net.upload_friend_image(user_id, image)
    
    wait_times = 90
    while not image.finish:
        await asyncio.sleep(1)
        wait_times -= 1
        if not wait_times:
            raise Exception(f"获取图片 [PQ:image,file={image.hash.hex().upper()}] 超时")

    writer.write_hex("B6 E8 AF AD E4 BD 93 E7 AE 80 00 00 06 01 43 02")
    writer.write_hex("00 1B")
    writer.write((utils.randstr(23) + ".jpg").encode())
    writer.write_hex("03 00 04")
    writer.write_int32(image.size)
    writer.write_hex("04")
    writer.write_int16(len(image.picid))
    writer.write(image.picid)
    writer.write_hex("14 00 04 11 00 00 00 0B 00 00 18")
    writer.write_int16(len(image.picid))
    writer.write(image.picid)
    writer.write_hex("19 00 04 00 00")
    writer.write_int16(image.width)
    writer.write_hex("1A 00 04 00 00")
    writer.write_int16(image.height)
    writer.write_hex("1F 00 04 00")

    writer.write_hex("00 03 E8 1B 00 10")
    writer.write(image.hash)
    writer.write_hex("FF 00 75 16")

    writer.write(b' 117101061CB')
    size = str(image.size).encode()
    if len(size) > 5:
        writer.write(b' ' * 4)
    else:
        writer.write(b' ' * 5)

    writer.write(size)
    writer.write_byte(ord("e"))
    writer.write((image.hash.hex().upper()+".jpg").encode())
    writer.write_byte(ord("x"))
    writer.write(image.picid)
    writer.write_byte(ord("A"))
    return writer.clear()



import sys
import logging


logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    datefmt="%Y/%m/%d-%H:%M:%S",
    format='[%(name)s %(asctime)s] %(levelname)s: %(message)s'
)

logger = logging.getLogger('PCQQ')
logger.fatal = sys.exit
logging.getLogger("asyncio").setLevel(logging.CRITICAL)

sys.modules[__name__] = logger



import re
import io
import struct
from xml.etree import ElementTree


def _convert_topx(value):
    matched = re.match(r"(\d+)(?:\.\d)?([a-z]*)$", value)
    if not matched:
        raise ValueError("unknown length value: %s" % value)
    else:
        length, unit = matched.groups()
        if unit == "":
            return int(length)
        elif unit == "cm":
            return int(length) * 96 / 2.54
        elif unit == "mm":
            return int(length) * 96 / 2.54 / 10
        elif unit == "in":
            return int(length) * 96
        elif unit == "pc":
            return int(length) * 96 / 6
        elif unit == "pt":
            return int(length) * 96 / 6
        elif unit == "px":
            return int(length)
        else:
            raise ValueError("unknown unit type: %s" % unit)


def img_size(im_data: bytes):
    """
    Return (width, height) for a given img file content
    no requirements
    :type filepath: Union[str, pathlib.Path]
    :rtype Tuple[int, int]
    """
    height = -1
    width = -1

    with io.BytesIO(im_data) as fhandle:
        head = fhandle.read(24)
        size = len(head)
        # handle GIFs
        if size >= 10 and head[:6] in (b'GIF87a', b'GIF89a'):
            # Check to see if content_type is correct
            try:
                width, height = struct.unpack("<hh", head[6:10])
            except struct.error:
                raise ValueError("Invalid GIF file")
        # see png edition spec bytes are below chunk length then and finally the
        elif size >= 24 and head.startswith(b'\211PNG\r\n\032\n') and head[12:16] == b'IHDR':
            try:
                width, height = struct.unpack(">LL", head[16:24])
            except struct.error:
                raise ValueError("Invalid PNG file")
        # Maybe this is for an older PNG version.
        elif size >= 16 and head.startswith(b'\211PNG\r\n\032\n'):
            # Check to see if we have the right content type
            try:
                width, height = struct.unpack(">LL", head[8:16])
            except struct.error:
                raise ValueError("Invalid PNG file")
        # handle JPEGs
        elif size >= 2 and head.startswith(b'\377\330'):
            try:
                fhandle.seek(0)  # Read 0xff next
                size = 2
                ftype = 0
                while not 0xc0 <= ftype <= 0xcf or ftype in [0xc4, 0xc8, 0xcc]:
                    fhandle.seek(size, 1)
                    byte = fhandle.read(1)
                    while ord(byte) == 0xff:
                        byte = fhandle.read(1)
                    ftype = ord(byte)
                    size = struct.unpack('>H', fhandle.read(2))[0] - 2
                # We are at a SOFn block
                fhandle.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack('>HH', fhandle.read(4))
            except struct.error:
                raise ValueError("Invalid JPEG file")
        # handle JPEG2000s
        elif size >= 12 and head.startswith(b'\x00\x00\x00\x0cjP  \r\n\x87\n'):
            fhandle.seek(48)
            try:
                height, width = struct.unpack('>LL', fhandle.read(8))
            except struct.error:
                raise ValueError("Invalid JPEG2000 file")
        # handle big endian TIFF
        elif size >= 8 and head.startswith(b"\x4d\x4d\x00\x2a"):
            offset = struct.unpack('>L', head[4:8])[0]
            fhandle.seek(offset)
            ifdsize = struct.unpack(">H", fhandle.read(2))[0]
            for i in range(ifdsize):
                tag, datatype, count, data = struct.unpack(
                    ">HHLL", fhandle.read(12))
                if tag == 256:
                    if datatype == 3:
                        width = int(data / 65536)
                    elif datatype == 4:
                        width = data
                    else:
                        raise ValueError(
                            "Invalid TIFF file: width column data type should be SHORT/LONG.")
                elif tag == 257:
                    if datatype == 3:
                        height = int(data / 65536)
                    elif datatype == 4:
                        height = data
                    else:
                        raise ValueError(
                            "Invalid TIFF file: height column data type should be SHORT/LONG.")
                if width != -1 and height != -1:
                    break
            if width == -1 or height == -1:
                raise ValueError(
                    "Invalid TIFF file: width and/or height IDS entries are missing.")
        elif size >= 8 and head.startswith(b"\x49\x49\x2a\x00"):
            offset = struct.unpack('<L', head[4:8])[0]
            fhandle.seek(offset)
            ifdsize = struct.unpack("<H", fhandle.read(2))[0]
            for i in range(ifdsize):
                tag, datatype, count, data = struct.unpack(
                    "<HHLL", fhandle.read(12))
                if tag == 256:
                    width = data
                elif tag == 257:
                    height = data
                if width != -1 and height != -1:
                    break
            if width == -1 or height == -1:
                raise ValueError(
                    "Invalid TIFF file: width and/or height IDS entries are missing.")
        # handle SVGs
        elif size >= 5 and head.startswith(b'<?xml'):
            try:
                fhandle.seek(0)
                root = ElementTree.parse(fhandle).getroot()
                width = _convert_topx(root.attrib["width"])
                height = _convert_topx(root.attrib["height"])
            except Exception:
                raise ValueError("Invalid SVG file")

    return width, height



import io
import re


class Reader:
    def __init__(self, initial_bytes: bytes = b'') -> None:
        self.raw = io.BytesIO(initial_bytes)

    def tell(self) -> int:
        return len(self.raw.getvalue()) - self.raw.tell()

    def read(self, size: int = -1) -> bytes:
        return self.raw.read(size)

    def read_hex(self, size: int = -1) -> str:
        hex = self.read(size).hex()
        return " ".join(re.findall('.{2}', hex.upper()))

    def read_int16(self) -> int:
        return int.from_bytes(self.read(2), 'big')

    def read_int32(self) -> int:
        return int.from_bytes(self.read(4), 'big')

    def read_byte(self) -> int:
        return self.read(1)[0]



"""

"""
from .bot import only_login, run_bot
from .network import (
    login_out,
    set_group_card,
    set_group_shutup,
    send_friend_msg,
    send_group_msg,
    set_online,
)
from .plugin import (
    on,
    on_full,
    on_fulls,
    on_keyword,
    on_keywords,
    on_command,
    on_commands,
    on_regex,

    only_group,
    only_friend,
    check_at,
    check_type,
    check_user,
    check_group,
    check_session,
    must_given
)
from .session import Session



import os
import asyncio
import platform
import subprocess

import pcqq.client as cli
import pcqq.utils as utils
import pcqq.const as const
import pcqq.logger as logger
import pcqq.binary as binary
#cli = cli.QQClient()


async def ping_server(is_logging: bool):
    """
    0825包 一切的开始  均衡负载

    发这个包就问候服务器 Say Hello 判断是否能链接服务器
    """
    writer = binary.Writer()

    writer.write(binary.tlv018(cli.uin, cli.redirection_times))
    if is_logging:
        writer.write(binary.tlv309(
            cli.server_ip,
            cli.redirection_history,
            cli.redirection_times
        ))
        writer.write(binary.tlv036())
    else:
        writer.write(binary.tlv004())
        writer.write(binary.tlv309(
            cli.server_ip,
            cli.redirection_history,
            cli.redirection_times
        ))
        writer.write(binary.tlv114())

    tea = binary.QQTea(bytes.fromhex(const.RANDKEY))
    await cli.write_packet(
        "08 25",
        const.STRUCT_VERSION,
        bytes.fromhex(const.RANDKEY) +
        tea.encrypt(writer.clear())
    )

    _, _, _, body = await cli.read_packet()
    reader = binary.Reader(tea.decrypt(body))

    sign = reader.read_byte()
    reader.read(2)
    cli.token_0038_from_0825 = reader.read(reader.read_int16())
    reader.read(6)
    cli.login_time = reader.read(4)
    cli.local_ip = reader.read(4)
    reader.read(2)

    if sign == 0xfe:
        reader.read(18)
        cli.server_ip = reader.read(4)
        cli.redirection_times += 1
        cli.redirection_history += cli.server_ip
    elif sign == 0x00:
        reader.read(6)
        cli.server_ip = reader.read(4)
        cli.redirection_history = b''


async def fetch_qrcode():
    """
    0818包 向服务器申请登录二维码

    发这个包就代表需要使用手机扫描二维码登录
    """
    writer = binary.Writer()

    writer.write(binary.tlv019())
    writer.write(binary.tlv114())
    writer.write(binary.tlv305())

    tea = binary.QQTea(bytes.fromhex(const.RANDKEY))
    await cli.write_packet(
        "08 18",
        const.STRUCT_VERSION,
        bytes.fromhex(const.RANDKEY) +
        tea.encrypt(writer.clear())
    )

    _, _, _, body = await cli.read_packet()
    tea.key = bytes.fromhex(const.SHAREKEY)
    reader = binary.Reader(tea.decrypt(body))

    if reader.read_byte() != 0x00:
        logger.fatal("ERROR: 登录二维码获取失败，请尝试重新运行")

    reader.read(6)
    cli.pckey_for_0819 = reader.read(16)
    reader.read(4)
    cli.token_0038_from_0818 = reader.read(reader.read_int16())
    reader.read(4)
    cli.token_by_scancode = reader.read(reader.read_int16())
    reader.read(4)

    path = os.path.join(os.getcwd(), 'QrCode.jpg')
    with open(path, mode="wb") as f:
        f.write(reader.read(reader.read_int16()))
        logger.info('登录二维码获取成功，已保存至' + path)

    system = platform.system()
    try:
        # 用系统默认程序打开图片文件
        if system == "Windows":
            os.startfile(path)
        elif system == "Linux":
            subprocess.call(["xdg-open", path])
        else:
            subprocess.call(["open", path])
    except:
        # 使用pillow库在终端上打印二维码图片
        utils.print_qrcode(path, const.QRCODE_SIZE)


async def check_qrcode():
    """
    0819包 检查当前扫码状态
    """
    writer = binary.Writer()

    writer.write(binary.tlv019())
    writer.write(binary.tlv114())
    writer.write_hex("03 01 00 22")
    writer.write_int16(len(cli.token_by_scancode))
    writer.write(cli.token_by_scancode)

    tea = binary.QQTea(cli.pckey_for_0819)
    await cli.write_packet(
        "08 19",
        const.STRUCT_VERSION + "00 30 00 3A",
        len(cli.token_0038_from_0818).to_bytes(2, 'big') +
        cli.token_0038_from_0818 +
        tea.encrypt(writer.clear())
    )

    _, _, uin, body = await cli.read_packet()
    reader = binary.Reader(tea.decrypt(body))
    state = reader.read_byte()

    if state == 0x01:
        logger.info(f'账号 {uin} 已扫码，请在手机上确认登录')
    elif state == 0x00:
        cli.uin = uin

        reader.read(2)
        cli.password = reader.read(reader.read_int16())
        reader.read(2)
        cli.tgt_key = reader.read(reader.read_int16())

        os.remove('QrCode.jpg')
        logger.info(f'账号 {cli.uin} 已在手机上确认登录，尝试登录中...')
        return True
    return False


async def check_login():
    """
    0836包 登录验证/账密验证
    """
    writer = binary.Writer()

    writer.write(binary.tlv112(cli.token_0038_from_0825))
    writer.write(binary.tlv30f())
    writer.write(binary.tlv005(cli.uin))

    if cli.is_scancode:
        writer.write(binary.tlv303(cli.password))
    else:
        writer.write(binary.tlv006(
            uin=cli.uin,
            tgt_key=cli.tgt_key,
            md5_once=cli.password,
            md5_twice=utils.hashmd5(
                cli.password +
                bytes(4) +
                cli.uin.to_bytes(4, "big")
            ),
            login_time=cli.login_time,
            local_ip=cli.local_ip,
            computer_id=utils.hashmd5(
                f'{cli.uin}ComputerID'.encode()
            ),
            tgtgt=cli.tgtgt_key
        ))

    writer.write(binary.tlv015())
    writer.write(binary.tlv01a(cli.tgt_key))
    writer.write(binary.tlv018(cli.uin, cli.redirection_times))
    writer.write(binary.tlv103())
    writer.write(binary.tlv312())
    writer.write(binary.tlv508())
    writer.write(binary.tlv313())
    writer.write(binary.tlv102(cli.token_0038_from_0825))

    tea = binary.QQTea(bytes.fromhex(const.SHAREKEY))
    await cli.write_packet(
        "08 36",
        const.STRUCT_VERSION + "00 01 01 02",
        len(bytes.fromhex(const.PUBLICKEY)).to_bytes(2, 'big') +
        bytes.fromhex(const.PUBLICKEY) +
        bytes.fromhex("00 00 00 10") +
        bytes.fromhex(const.RANDKEY) +
        tea.encrypt(writer.clear())
    )

    try:
        _, _, _, body = await cli.read_packet()
    except:
        logger.fatal("ERROR: "+", ".join([
            "登录验证失败",
            "可能是您的设备开启了登录保护",
            "请在手机QQ的[设置]->[账号安全]->[登录设备管理]中关闭[登录保护]选项"
        ]))

    try:
        body = tea.decrypt(body)
        tea.key = cli.tgt_key
        reader = binary.Reader(tea.decrypt(body))
    except:
        logger.fatal("ERROR: 登录失败，请尝试重启程序或使用扫码登录")

    sign = reader.read_byte()
    if sign == 0x00:
        for _ in range(5):
            reader.read_byte()

            binary.tlv_id = reader.read_byte()
            binary.tlv_size = reader.read_int16()

            if binary.tlv_id == 9:
                reader.read(2)
                cli.pckey_for_0828_send = reader.read(16)

                cli.token_0038_from_0836 = reader.read(
                    reader.read_int16())

                reader.read(reader.read_int16())
                reader.read(2)
            elif binary.tlv_id == 8:
                reader.read(8)
                cli.nickname = reader.read(reader.read_byte()).decode()
            elif binary.tlv_id == 7:
                reader.read(26)
                cli.pckey_for_0828_recv = reader.read(16)

                cli.token_0088_from_0836 = reader.read(
                    reader.read_int16())
                reader.read(binary.tlv_size - 180)
            else:
                reader.read(binary.tlv_size)
    elif sign == 0x01:
        reader.read(2)
        cli.tgt_key = reader.read(reader.read_int16())
        reader.read(2)
        cli.tgtgt_key = reader.read(reader.read_int16())

        tea.key = utils.hashmd5(
            cli.password +
            bytes(4) +
            cli.uin.to_bytes(4, "big")
        )
        cli.tgtgt_key = tea.decrypt(cli.tgtgt_key)

        return await check_login()
    else:
        logger.fatal(f"ERROR: 登录失败，错误码{hex(sign)}，请尝试重新运行或使用扫码登录")


async def fetch_session():
    """
    0828包 申请SessionKey
    此包代表获取会话密钥来操作协议功能
    """
    writer = binary.Writer()

    writer.write(binary.tlv007(cli.token_0088_from_0836))
    writer.write(binary.tlv00c(cli.server_ip))
    writer.write(binary.tlv015())
    writer.write(binary.tlv036())
    writer.write(binary.tlv018(cli.uin, cli.redirection_times))
    writer.write(binary.tlv01f())
    writer.write(binary.tlv105())
    writer.write(binary.tlv10b())
    writer.write(binary.tlv02d(cli.local_ip))

    tea = binary.QQTea(cli.pckey_for_0828_send)
    await cli.write_packet(
        "08 28",
        const.BODY_VERSION + "00 30 00 3A 00 38",
        cli.token_0038_from_0836 +
        tea.encrypt(writer.clear())
    )

    _, _, _, body = await cli.read_packet()
    tea.key = cli.pckey_for_0828_recv
    reader = binary.Reader(tea.decrypt(body))
    reader.read(63)
    cli.session_key = reader.read(16)
    cli.teaer = binary.QQTea(cli.session_key)


async def set_online(state: int):
    """00EC包 设置QQ状态"""
    writer = binary.Writer()

    writer.write_hex("01 00")
    writer.write_byte(state)
    writer.write_hex("00 01")
    writer.write_hex("00 01")
    writer.write_hex("00 04")
    writer.write_hex("00 00 00 00")

    await cli.write_packet(
        "00 EC",
        const.BODY_VERSION,
        writer.clear()
    )
    await cli.read_packet()
    logger.info(f"账号 {cli.uin} 已上线，欢迎用户 {cli.nickname} 使用本协议库")


async def login_out():
    """0062包 注销当前登录"""
    await cli.write_packet(
        "00 62",
        const.BODY_VERSION,
        bytes(16)
    )
    logger.info(f"账号 {cli.uin} 已退出登录，欢迎用户 {cli.nickname} 下次使用本协议库")


async def heatbeat():
    """0058包 与服务端保持心跳"""
    await cli.write_packet(
        "00 58",
        const.BODY_VERSION,
        bytes.fromhex("00 01 00 01")
    )


async def fetch_skey():
    """001D包 获取网络操作用的skey"""
    writer = binary.Writer()

    writer.write_byte(51)
    writer.write_int16(6)   # 域名数量

    writer.write_int16(8)
    writer.write(b't.qq.com')

    writer.write_int16(10)
    writer.write(b'qun.qq.com')

    writer.write_int16(12)
    writer.write(b'qzone.qq.com')

    writer.write_int16(12)
    writer.write(b'jubao.qq.com')

    writer.write_int16(9)
    writer.write(b'ke.qq.com')

    writer.write_int16(10)
    writer.write(b'tenpay.com')

    await cli.write_packet(
        "00 1D",
        const.BODY_VERSION,
        writer.clear()
    )

    _, _, _, body = await cli.read_packet()
    reader = binary.Reader(body)

    reader.read(2)
    cli.skey = reader.read(reader.read_int16()).decode()
    cli.bkn = utils.gtk_skey(cli.skey)
    cli.cookie = "uin=o%d; skey=%s" % (cli.uin, cli.skey)

async def qrcode_login():
    """通过手机扫码登录QQ"""
    cli.is_scancode = True
    await ping_server(False)  # 请求登录服务器
    await fetch_qrcode()    # 获取登录二维码

    # 循环检测扫码状态
    while not await check_qrcode():
        await asyncio.sleep(1.0)
    
    await check_login() # 验证登录
    await fetch_session()   # 申请会话操作密钥
    await set_online(const.STATE_ONLINE)    # 设置状态为在线
    await fetch_skey()    # 申请网络操作密钥

    cli.save_token(os.path.join(os.getcwd(), "session.token"))

async def password_login(user_id:int, password:str):
    """通过账号密码登录QQ"""
    cli.uin = user_id
    cli.password = utils.hashmd5(password.encode())

    await ping_server(False)  # 请求登录服务器
    await check_login() # 账密登录
    await fetch_session()   # 申请会话操作密钥
    await set_online(const.STATE_ONLINE)    # 设置状态为在线
    await fetch_skey()    # 申请网络操作密钥

    cli.save_token(os.path.join(os.getcwd(), "session.token"))

async def token_login():
    """通过本地token申请重连"""
    await ping_server(False)  # 请求登录服务器

    cli.load_token(os.path.join(os.getcwd(), "session.token"))
    await ping_server(False)  # 重新请求登录服务器
    
    try:
        await fetch_session() # 验证登录
    except:
        os.remove(os.path.join(os.getcwd(), "session.token"))
        logger.fatal("ERROR: 本地Token已失效，请重新登录")
    
    await set_online(const.STATE_ONLINE)    # 设置状态为在线
    await fetch_skey()    # 申请网络操作密钥


import os
import asyncio

import pcqq.handle as hd
import pcqq.client as cli
import pcqq.const as const
import pcqq.network as net
import pcqq.session as ses
import pcqq.logger as logger
# cli = cli.QQClient()


async def exec_rules(rules, session):
    # 判断是否执行插件
    for rule in rules:
        if not await rule(session):
            return False
    return True


async def plugin_handle(body):
    session = ses.Session()
    typ = body[18:20].hex().upper()
    if typ == const.EVENT_GROUP_MSG:
        await hd.group_msg_handle(session, body)
    elif typ == const.EVENT_FRIEND_MSG:
        await hd.friend_msg_handle(session, body)
    elif typ == const.EVENT_GROUP_INCREASE:
        await hd.group_increase_handle(session, body)
    elif typ == const.EVENT_GROUP_DECREASE:
        await hd.group_decrease_handle(session, body)
    elif typ == const.EVENT_OTHRE:
        await hd.other_handle(session, body)
    else:
        return

    if session.user_id == session.self_id and session.event_type == "group_msg":
        return  # 忽略自身群消息

    for plugin in cli.plugins:

        # 判断是否位临时插件
        if plugin.temp:
            del cli.plugins[cli.plugins.index(plugin)]

        # 判断是否执行插件
        if await exec_rules(plugin.rules, session):
            try:
                await plugin.handle(session)
            except Exception as err:
                logger.error(f"插件 {plugin.handle.__name__} 运行时发生异常 -> {err}")

        # 判断是否阻断后续插件执行
        if plugin.block:
            break

async def event_handle(cmd:str, sequence:bytes, body:bytes):
    await cli.write_packet(
        cmd,
        const.BODY_VERSION,
        body[0:16],
        sequence=sequence
    )   # 初步进行回执

    if cmd == "03 52":
        """私聊图片上传响应"""
        callback = cli.waiter.pop(sequence.hex())
        await callback(body)
    elif cmd == "03 88":
        """群聊图片上传响应"""
        callback = cli.waiter.pop(sequence.hex())
        await callback(body)
    else:
        await plugin_handle(body)

async def main_handle():
    while True:
        try:
            cmd, sequence, _, body = await cli.read_packet()
        except:
            continue
        cli.run_future(event_handle(cmd, sequence, body))

async def keep_heatbeat():
    while True:
        cli.run_future(net.heatbeat())
        await asyncio.sleep(40.0)


def only_login(uin: int = 0, password: str = ""):
    """
    仅登录, 不处理事件

    :param uin: 用于登录的QQ号(留空则使用扫码登录)

    :param password: 登录QQ号的密码(留空则使用扫码登录)

    """
    cli.run(cli.client.init())

    if os.path.exists("session.token"):
        cli.run(net.token_login())
    elif uin and password:
        cli.run(net.password_login(uin, password))
    else:
        cli.run(net.qrcode_login())


def run_bot(uin: int = 0, password: str = ""):
    """
    登录机器人,并开始处理事件

    :param uin: 用于登录的QQ号(留空则使用扫码登录)

    :param password: 登录QQ号的密码(留空则使用扫码登录)

    """
    only_login(uin, password)
    cli.run_future(keep_heatbeat())

    cli.run_future(main_handle())
    logger.info(f"已导入 {len(cli.plugins)} 组插件，开始处理事件.....")
    cli.client.loop.run_forever()



import time
import random
import string
import hashlib

def hashmd5(src: bytes):
    return hashlib.md5(src).digest()

def randbytes(size: int):
    return bytes([random.randint(0, 255) for _ in range(size)])

def randstr(size: int):
    return ''.join(random.sample(string.ascii_letters + string.digits, size))


def gtk_skey(skey: str) -> int:
    accu = 5381
    for s in skey:
        accu += (accu << 5) + ord(s)
    return accu & 2147483647


def time_lapse(seconds: int):
    time_stamp = int(time.time()) + seconds
    return time.strftime("%Y年%m月%d日 %H:%M:%S", time.localtime(time_stamp))


import pcqq.logger as logger

def get_char(r, g, b, alpha=256):
    if alpha == 0:
        return ' '
    gary = (2126 * r + 7152 * g + 722 * b) / 10000
    ascii_char = list("■□")
    # gary / 256 = x / len(ascii_char)
    x = int(gary / (alpha + 1.0) * len(ascii_char))
    return ascii_char[x]


def make_qrtex(QR_Tab):
    tmp_text = ""
    print_tex = ""

    atrr = 7
    fore = 37
    back = 47
    color_block = "\x1B[%d;%d;%dm" % (atrr, fore, back)
    atrr = 0
    fore = 0
    back = 0
    color_none = "\x1B[%d;%d;%dm" % (atrr, fore, back)

    for loop in range(0, len(QR_Tab)):
        if QR_Tab[loop] == '■':
            tmp_text = "%s  \x1B[0m" % (color_block)
        elif QR_Tab[loop] == '□':
            tmp_text = "%s  \x1B[0m" % (color_none)
        else:
            tmp_text = "\n"

        print_tex = print_tex + tmp_text

    return print_tex


def print_qrcode(path:str, length: int):
    try:
        from PIL import Image
    except:
        logger.error("无法在终端输出二维码，请先安装 pillow 库")
        return

    im = Image.open(path)
    im = im.resize((length, length), Image.NEAREST)

    txt = ''
    for i in range(length):
        for j in range(length):
            content = im.getpixel((j, i))
            if isinstance(content, int):
                content = (content, content, content)
            txt += get_char(*content)
        txt += '\n'
    im.close()
    print(make_qrtex(txt))



def gid_from_group(group_id: int) -> int:
    group = str(group_id)
    left = int(group[0:-6])
    if left >= 0 and left <= 10:
        right = group[-6:]
        gid = str(left + 202) + right
    elif left >= 11 and left <= 19:
        right = group[-6:]
        gid = str(left + 469) + right
    elif left >= 20 and left <= 66:
        left = int(str(left)[0:1])
        right = group[-7:]
        gid = str(left + 208) + right
    elif left >= 67 and left <= 156:
        right = group[-6:]
        gid = str(left + 1943) + right
    elif left >= 157 and left <= 209:
        left = int(str(left)[0:2])
        right = group[-7:]
        gid = str(left + 199) + right
    elif left >= 210 and left <= 309:
        left = int(str(left)[0:2])
        right = group[-7:]
        gid = str(left + 389) + right
    elif left >= 310 and left <= 335:
        left = int(str(left)[0:2])
        right = group[-7:]
        gid = str(left + 349) + right
    elif left >= 336 and left <= 386:
        left = int(str(left)[0:3])
        right = group[-6:]
        gid = str(left + 2265) + right
    elif left >= 387 and left <= 499:
        left = int(str(left)[0:3])
        right = group[-6:]
        gid = str(left + 3490) + right
    elif left >= 500:
        return int(group)
    return int(gid)


def group_from_gid(gid: int) -> int:
    gid = str(gid)
    if int(gid[0:3]) >= 500:
        return int(gid)
    left = int(gid[0:-6])

    if left == 202:
        right = gid[-6:]
        group = int(right)
    elif left >= 203 and left <= 212:
        right = gid[-6:]
        group = int(str(left - 202) + right)
    elif left >= 480 and left <= 488:
        right = gid[-6:]
        group = int(str(left - 469) + right)
    elif left >= 2010 and left <= 2099:
        right = gid[-6:]
        group = int(str(left - 1943) + right)
    elif left >= 2100 and left <= 2146:
        left = int(str(left)[0:3])
        right = gid[-7:]
        group = int(str(left - 208) + right)
    elif left >= 2147 and left <= 2199:
        left = int(str(left)[0:3])
        right = gid[-7:]
        group = int(str(left - 199) + right)
    elif left >= 2601 and left <= 2651:
        left = int(str(left)[0:4])
        right = gid[-6:]
        group = int(str(left - 2265) + right)
    elif left >= 3800 and left <= 3989:
        left = int(str(left)[0:3])
        right = gid[-7:]
        group = int(str(left - 349) + right)
    elif left >= 4100 and left <= 4199:
        left = int(str(left)[0:3])
        right = gid[-7:]
        group = int(str(left - 389) + right)
    else:
        group = 0

    return group


import json
import pcqq.utils as utils
import pcqq.client as cli
import pcqq.logger as logger


def write_ginfo(info: dict):
    if not cli.cache.exist("ginfo", group_id=info["gid"]):
        cli.cache.insert(
            "ginfo",
            info["name"],
            info["owner"],
            info["gid"],
            info["board"],
            info["intro"],
            info["mem_num"],
        )
    for mem in info["members"]:
        if cli.cache.exist("gmember", user_id=mem["uid"], group_id=info["gid"]):
            continue
        cli.cache.insert(
            "gmember",
            mem["card"],
            mem["uid"],
            mem["name"],
            info["gid"],
            mem["op"]
        )
    cli.cache.commit()


# 封装 Web Api 调用
# cli = cli.QQClient()

async def get_group_members(group_id: int) -> list:
    """
    获取群成员列表
    """
    rsp = await cli.webget(
        "https://qinfo.clt.qq.com/cgi-bin/qun_info/get_group_members_new",
        gc=group_id,
        bkn=cli.bkn,
        src="qinfo_v3",
        headers={"Cookie": cli.cookie}
    )
    rsp.content = rsp.content.split(b'\r\n')[1]
    rsp.content = rsp.content.replace(b"&nbsp;", b" ").replace(b"&amp;", b"&")
    ret = rsp.json()

    cards = ret.get("cards", {})
    admins = [ret["owner"], *ret.get("adm", [])]
    return [
        {
            "uid": mem["u"],
            "name": mem["n"],
            "card": cards.get(str(mem["u"]), mem["n"]),
            "op": int(mem["u"] in admins)
        }
        for mem in ret["mems"]
    ]


async def get_group_info(group_id: int) -> dict:
    """
    获取群信息
    """
    rsp = await cli.webget(
        "https://qinfo.clt.qq.com/cgi-bin/qun_info/get_group_info_all",
        gc=group_id,
        bkn=cli.bkn,
        src="qinfo_v3",
        headers={"Cookie": cli.cookie}
    )
    rsp.content = rsp.content.split(b'\r\n')[1]
    rsp.content = rsp.content.replace(b"&nbsp;", b" ").replace(b"&amp;", b"&")
    ret = rsp.json()

    return {
        "gid": group_id,
        "name": ret["gName"],
        "owner": ret["gOwner"],
        "board": ret["gBoard"],
        "intro": ret["gRIntro"],
        "mem_num": ret["gMemNum"],
        "members": await get_group_members(group_id)
    }


temp_cache = {}


async def get_user_name(user_id: int) -> str:
    if user_id in temp_cache:
        return temp_cache[user_id]

    rsp = await cli.webget("https://r.qzone.qq.com/fcg-bin/cgi_get_score.fcg", mask=7, uins=user_id)
    rsp.content = rsp.content[17:-2]
    temp_cache[user_id] = rsp.json()[str(user_id)][-4]

    return temp_cache[user_id]


async def set_group_card(group_id: int, user_id: int,  nickname: str) -> None:
    """
    修改群成员卡片

    :param group_id: 被修改成员所在的群号

    :param user_id: 被修改成员的QQ号

    :param nickname: 修改后的昵称

    """
    rsp = await cli.webpost(
        url="https://qinfo.clt.qq.com/cgi-bin/qun_info/set_group_card",
        u=user_id,
        name=nickname,
        gc=group_id,
        bkn=cli.bkn,
        src="qinfo_v3",
        headers={"Cookie": cli.cookie}
    )
    rsp.content = rsp.content.split(b'\r\n')[1]
    rsp.content = rsp.content.replace(b"&nbsp;", b" ").replace(b"&amp;", b"&")
    ret = rsp.json()

    gname = await get_group_cache(group_id)
    cord = await get_user_cache(user_id, group_id)
    if ret["em"] == "ok":
        logger.info("成功将 %s(%d) 在群聊 %s(%d) 的名片设置为 %s" % (
            cord, user_id,
            gname, group_id, nickname
        ))
    else:
        logger.error("无法在群聊 %s(%d) 中修改 %s(%d) 的名片" % (
            gname, group_id,
            cord, user_id
        ))


async def set_group_shutup(group_id: int, user_id: int, secs: int = 60):
    """
    设置群成员禁言

    :param group_id: 被禁言成员所在的群号

    :param user_id: 被禁言成员的QQ号

    :param secs: 禁言时长(单位: 秒)

    """
    rsp = await cli.webget(
        url="https://qinfo.clt.qq.com/cgi-bin/qun_info/set_group_shutup",
        shutup_list=json.dumps([{"uin": user_id, "t": secs}]),
        gc=group_id,
        bkn=cli.bkn,
        src="qinfo_v3",
        headers={"Cookie": cli.cookie,}
    )
    rsp.content = rsp.content.split(b'\r\n')[1]
    rsp.content = rsp.content.replace(b"&nbsp;", b" ").replace(b"&amp;", b"&")
    ret = rsp.json()

    gname = await get_group_cache(group_id)
    cord = await get_user_cache(user_id, group_id)
    if not ret["ec"]:

        logger.info("成功将 %s(%d) 在群聊 %s(%d) 中禁言， 预计解禁时间 %s" % (
            cord, user_id,
            gname, group_id,
            utils.time_lapse(secs)
        ))
    else:
        logger.error("无法将 %s(%d) 在群聊 %s(%d) 中禁言" % (
            cord, user_id,
            gname, group_id
        ))

# 封装 cache 简单调用


async def get_user_cache(user_id: int, group_id: int = 0) -> tuple:
    """
    在缓存内取用户昵称
    """
    if group_id:
        ret = cli.cache.select(
            table_name="gmember",
            user_id=user_id,
            group_id=group_id
        )
        if not ret:
            write_ginfo(await get_group_info(group_id))
            return await get_user_cache(user_id, group_id)
        return ret[0][0]
    else:
        return await get_user_name(user_id)


async def get_group_cache(group_id: int) -> str:
    """
    在缓存内取群名称
    """
    ret = cli.cache.select("ginfo", group_id=group_id)
    if not ret:
        write_ginfo(await get_group_info(group_id))
        return await get_group_cache(group_id)
    return ret[0][0]



from .sqlite import SqliteDB
from .imsize import img_size
from .qrcode import print_qrcode
from .group import (
    gid_from_group,
    group_from_gid
)
from .helper import (
    hashmd5,
    randstr,
    randbytes,
    gtk_skey,
    time_lapse
)



import re
import asyncio
import pcqq.client as cli
import pcqq.logger as logger


class BasePlugin:
    def __init__(self, rules):
        self.rules = rules
        self.temp = ...
        self.block = ...
        self.handle = ...
        self.priority = ...

    def __call__(self, handle):
        self.handle = handle
        return self

    def SetTemp(self, temp: bool):
        self.temp = temp
        return self

    def SetBlock(self, block: bool):
        self.block = block
        return self

    def SetPriority(self, priority=10):
        self.priority = priority
        return self

# Builtin Plugin Rule


async def only_group(session):
    return bool(session.group_id)


async def only_friend(session):
    return not session.group_id and session.user_id


def check_at(user_id):
    async def check_rule(session):
        return {"type": "at", "qq": user_id} in session.raw_message
    return check_rule


def check_type(*type_group):
    async def check_rule(session):
        return session.event_type in type_group
    return check_rule


def check_user(*uid_group):
    async def check_rule(session):
        return session.user_id in uid_group
    return check_rule


def check_group(*gid_group):
    async def check_rule(session):
        return session.group_id in gid_group
    return check_rule


def check_session(session_):
    async def check_rule(session):
        user_eq = session_.user_id == session.user_id
        group_eq = session_.group_id == session.group_id
        time_eq = session_.timestamp < session.timestamp
        return user_eq and group_eq and time_eq
    return check_rule


def must_given(prompt):
    async def wait_rule(session):
        await session.send_msg({"type": "at", "qq": session.user_id}, prompt)

        @on(check_session(session), temp=True, block=True, priority=-float("inf"))
        async def wait_input(next_session):
            session.matched = next_session.message

        wait_times = 60
        while not session.matched and wait_times:
            await asyncio.sleep(0.5)
            wait_times -= 1

        try:
            del cli.plugins[cli.plugins.index(wait_input)]
            logger.error(f"等待会话 {session.user_id} 输入 -> 超时")
        except ValueError:
            logger.info(f"捕获会话 {session.user_id} 输入 -> {session.matched}")

            return bool(session.matched)
    return wait_rule

# Builtin Plugin registrar


def on(*rules, temp=False, block=False, priority=10):
    plugin = BasePlugin(rules)
    plugin.SetTemp(temp)
    plugin.SetBlock(block)
    plugin.SetPriority(priority)

    cli.plugins.append(plugin)
    cli.plugins.sort(key=lambda p: p.priority)
    return plugin


def on_full(key, *rules, **params):
    async def full_rule(session):
        return key == session.message
    return on(*rules, full_rule, **params)


def on_fulls(key_group, *rules, **params):
    async def fulls_rule(session):
        return session.message in key_group
    return on(*rules, fulls_rule, **params)


def on_keyword(key, *rules, **params):
    async def key_rule(session):
        return key in session.message
    return on(*rules, key_rule, **params)


def on_keywords(key_group, *rules, **params):
    async def keys_rule(session):
        for key in key_group:
            if key in session.message:
                return True
        return False
    return on(*rules, keys_rule, **params)


def on_command(cmd, *rules, prompt="", **params):
    async def cmd_rule(session):
        if session.message.startswith(cmd):
            session.matched = session.message[len(cmd):].strip()

        if session.matched == "" and prompt:
            return await must_given(prompt)(session)

        return bool(session.matched)

    return on(*rules, cmd_rule, **params)


def on_commands(cmd_group, *rules, prompt="", **params):
    async def cmds_rule(session):
        if session.matched:
            return True

        for cmd in cmd_group:
            if session.message.startswith(cmd):
                session.matched = session.message[len(cmd):].strip()

            if session.matched == "" and prompt:
                return await must_given(prompt)(session)

        return bool(session.matched)

    return on(*rules, cmds_rule, **params)


def on_regex(pattern, *rules, **params):
    async def reg_rule(session):
        session.matched = re.findall(pattern, session.message)

        return bool(session.matched)
    return on(*rules, reg_rule, **params)



import struct


class QQTea:
    def __init__(self, key: bytes) -> None:
        self.key = key

    def encrypt(self, src: bytes) -> bytes:
        END_CHAR = b'\0'
        FILL_N_OR = 0xF8
        vl = len(src)
        filln = (8 - (vl + 2)) % 8 + 2
        fills = b''
        for i in range(filln):
            fills = fills + bytes([220])
        src = (bytes([(filln - 2) | FILL_N_OR])
               + fills
               + src
               + END_CHAR * 7)
        tr = b'\0' * 8
        to = b'\0' * 8
        r = b''
        o = b'\0' * 8
        for i in range(0, len(src), 8):
            o = xor(src[i:i + 8], tr)
            tr = xor(code(o, self.key), to)
            to = o
            r += tr
        return r

    def decrypt(self, src: bytes) -> bytes:
        l = len(src)
        prePlain = decipher(src, self.key)
        pos = (prePlain[0] & 0x07) + 2
        r = prePlain
        preCrypt = src[0:8]
        for i in range(8, l, 8):
            x = xor(decipher(xor(src[i:i + 8], prePlain), self.key), preCrypt)
            prePlain = xor(x, preCrypt)
            preCrypt = src[i:i + 8]
            r += x
        if r[-7:] != b'\0' * 7:
            return None
        return r[pos + 1:-7]


def xor(a, b) -> bytes:
    op = 0xffffffff
    a1, a2 = struct.unpack(b'>LL', a[0:8])
    b1, b2 = struct.unpack(b'>LL', b[0:8])
    return struct.pack(b'>LL', (a1 ^ b1) & op, (a2 ^ b2) & op)


def code(v, k) -> bytes:
    n = 16
    op = 0xffffffff
    delta = 0x9e3779b9
    k = struct.unpack(b'>LLLL', k[0:16])
    y, z = struct.unpack(b'>LL', v[0:8])
    s = 0
    for _ in range(n):
        s += delta
        y += (op & (z << 4)) + k[0] ^ z + s ^ (op & (z >> 5)) + k[1]
        y &= op
        z += (op & (y << 4)) + k[2] ^ y + s ^ (op & (y >> 5)) + k[3]
        z &= op
        r = struct.pack(b'>LL', y, z)
    return r


def decipher(v, k):
    n = 16
    op = 0xffffffff
    y, z = struct.unpack(b'>LL', v[0:8])
    a, b, c, d = struct.unpack(b'>LLLL', k[0:16])
    delta = 0x9E3779B9
    s = (delta << 4) & op
    for _ in range(n):
        z -= ((y << 4) + c) ^ (y + s) ^ ((y >> 5) + d)
        z &= op
        y -= ((z << 4) + a) ^ (z + s) ^ ((z >> 5) + b)
        y &= op
        s -= delta
        s &= op
    return struct.pack(b'>LL', y, z)



import pcqq.network as net
import pcqq.utils as utils
import pcqq.const as const
import pcqq.logger as logger
import pcqq.binary as binary


async def msg_handle(session, reader: binary.Reader):
    reader.read(16)
    reader.read(reader.read_int16())
    reader.read_int16()

    session.raw_message.clear()
    while reader.tell() > 0:
        typ = reader.read_byte()
        msgread = binary.Reader(reader.read(reader.read_int16()))

        if typ == const.MSG_TEXT:
            msgread.read_byte()
            data = msgread.read(msgread.read_int16())

            if msgread.tell():
                msgread.read(10)
                uid = msgread.read_int32()
                session.message += f"[PQ:at,qq={uid}]"
                session.raw_message.append({
                    "type": "at",
                    "qq": uid
                })
            else:
                text = data.decode()
                session.message += text
                session.raw_message.append(text)

        elif typ == const.MSG_FACE:
            msgread.read(3)
            face_id = msgread.read_byte()
            session.message += f"[PQ:face,id={face_id}]"
            session.raw_message.append({
                "type": "face",
                "id": face_id
            })

        elif typ in (const.MSG_IMAGE_GROUP, const.MSG_IMAGE_FRIEND):
            msgread.read_byte()
            id = msgread.read(msgread.read_int16()).decode().upper()
            id = id[:-4].replace("-", "").replace("{", "").replace("}", "")
            session.message += f"[PQ:image,file={id}]"
            session.raw_message.append({
                "type": "image",
                "url": f"https://gchat.qpic.cn/gchatpic_new/0/0-0-{id}/0?term=3"
            })


async def group_msg_handle(session, body: bytes):
    reader = binary.Reader(body)
    session.event_type = "group_msg"

    reader.read(4)
    session.self_id = reader.read_int32()

    reader.read(12)
    reader.read(reader.read_int32())
    session.group_id = reader.read_int32()

    reader.read_byte()
    session.user_id = reader.read_int32()
    session.msg_id = reader.read_int32()
    session.timestamp = reader.read_int32()

    reader.read(8)
    reader.read(16)
    await msg_handle(session, reader)

    if session.user_id != session.self_id:
        logger.info(f"收到群聊 %s(%d) 消息 %s(%d): %s" % (
            await net.get_group_cache(session.group_id),
            session.group_id,
            await net.get_user_cache(session.user_id, session.group_id),
            session.user_id,
            session.message
        ))
    await net.group_receipt(session.group_id, session.msg_id)


async def friend_msg_handle(session, body: bytes):
    reader = binary.Reader(body)
    session.event_type = "friend_msg"

    session.user_id = reader.read_int32()
    session.self_id = reader.read_int32()

    reader.read(12)
    reader.read(reader.read_int32())
    reader.read(26 + 2)

    session.msg_id = reader.read_int16()
    session.timestamp = reader.read_int32()

    reader.read(6 + 4 + 9)
    await msg_handle(session, reader)

    logger.info(f"收到好友消息 %s(%d): %s" % (
        await net.get_user_cache(session.user_id),
        session.user_id, session.message
    ))
    await net.friend_receipt(session.user_id, session.timestamp)


async def group_increase_handle(session, body: bytes):
    reader = binary.Reader(body)
    session.event_type = "group_increase"

    session.group_id = utils.group_from_gid(reader.read_int32())
    session.self_id = reader.read_int32()

    reader.read(12)
    reader.read(reader.read_int32() + 5)
    session.target_id = reader.read_int32()
    sign = reader.read_byte()
    session.user_id = reader.read_int32()

    if sign == 0x03:
        logger.info("收到群 %s(%d) 事件: %s(%d) 邀请 %s(%d) 加入群聊" % (
            await net.get_group_cache(session.group_id), session.group_id,
            await net.get_user_cache(
                session.user_id,
                session.group_id
            ), session.user_id,
            await net.get_user_cache(session.target_id), session.target_id
        ))

    else:
        logger.info("收到群 %s(%d) 事件: %s(%d) 同意 %s(%d) 加入群聊" % (
            await net.get_group_cache(session.group_id), session.group_id,
            await net.get_user_cache(
                session.user_id,
                session.group_id
            ), session.user_id,

            await net.get_user_cache(session.target_id), session.target_id
        ))


async def group_decrease_handle(session, body: bytes):
    reader = binary.Reader(body)
    session.event_type = "group_decrease"

    session.group_id = utils.group_from_gid(reader.read_int32())
    session.self_id = reader.read_int32()

    reader.read(12)
    reader.read(reader.read_int32() + 5)
    session.target_id = reader.read_int32()
    sign = reader.read_byte()
    session.user_id = reader.read_int32()

    if sign == 0x01:
        logger.info("收到群 %s(%d) 事件: 群主 %s(%d) 解散了该群" % (
            await net.get_group_cache(session.group_id), session.group_id,

            await net.get_user_cache(
                session.user_id,
                session.group_id
            ), session.user_id,
        ))
    elif sign == 0x02:
        logger.info("收到群 %s(%d) 事件: %s(%d) 退出了群聊" % (
            await net.get_group_cache(session.group_id), session.group_id,

            await net.get_user_cache(
                session.target_id,
                session.group_id
            ), session.target_id
        ))
    elif sign == 0x03:
        logger.info("收到群 %s(%d) 事件: %s(%d) 将 %s(%d) 踢出了群聊" % (
            await net.get_group_cache(session.group_id), session.group_id,

            await net.get_user_cache(
                session.user_id,
                session.group_id
            ), session.user_id,

            await net.get_user_cache(
                session.target_id,
                session.group_id
            ), session.target_id
        ))


async def shutup_handle(session, reader: binary.Reader):
    session.event_type = "group_shutup"

    reader.read(1)
    session.user_id = reader.read_int32()
    session.timestamp = reader.read_int32()

    reader.read(2)
    session.target_id = reader.read_int32()
    time = reader.read_int32()

    if session.target_id:
        if time:
            logger.info("收到群 %s(%d) 事件: %s(%d) 将 %s(%d) 禁言，预计 %s 禁言结束" % (
                await net.get_group_cache(session.group_id), session.group_id,

                await net.get_user_cache(
                    session.user_id,
                    session.group_id
                ), session.user_id,

                await net.get_user_cache(
                    session.target_id,
                    session.group_id
                ), session.target_id,

                utils.time_lapse(time)
            ))
        else:
            logger.info("收到群 %s(%d) 事件: %s(%d) 将 %s(%d) 解除禁言" % (
                await net.get_group_cache(session.group_id), session.group_id,

                await net.get_user_cache(
                    session.user_id,
                    session.group_id
                ), session.user_id,

                await net.get_user_cache(
                    session.target_id,
                    session.group_id
                ), session.target_id,
            ))
    else:
        if time:
            logger.info("收到群 %s(%d) 事件: %s(%d) 开启了全体禁言" % (
                await net.get_group_cache(session.group_id), session.group_id,

                await net.get_user_cache(
                    session.user_id,
                    session.group_id
                ), session.user_id,
            ))
        else:
            logger.info("收到群 %s(%d) 事件: %s(%d) 解除了全体禁言" % (
                await net.get_group_cache(session.group_id), session.group_id,

                await net.get_user_cache(
                    session.user_id,
                    session.group_id
                ), session.user_id,
            ))


async def anonymous_handle(session, reader: binary.Reader):
    session.event_type = "group_anonymous"
    pass


async def other_handle(session, body: bytes):
    reader = binary.Reader(body)
    session.group_id = utils.group_from_gid(reader.read_int32())
    session.self_id = reader.read_int32()

    reader.read(12)
    reader.read(reader.read_int32() + 4)
    sign = reader.read_byte()

    if sign == 0x0c:
        await shutup_handle(session, reader)
    elif sign == 0x0e:
        await anonymous_handle(session, reader)



MSG_TEXT = 0x01  # 纯文本
MSG_FACE = 0x02  # 表情
MSG_IMAGE_GROUP = 0X03  # 群聊图片
MSG_IMAGE_FRIEND = 0X06  # 私聊图片
MSG_XML = 0x14  # XML卡片
MSG_JSON = 0x25  # JSON卡片

STATE_ONLINE = 10   # 上线
STATE_LEAVE = 30    # 离开
STATE_INVISIBLE = 40    # 隐身
STATE_BUSY = 50  # 忙碌
STATE_CALLME = 60   # Q我吧
STATE_UNDISTURB = 70    # 请勿打扰

EVENT_GROUP_MSG = "0052"    # 群消息
EVENT_FRIEND_MSG = "00A6"   # 好友消息
EVENT_TEMP_MSG = "008D"  # 临时会话消息
EVENT_GROUP_INCREASE = "0021"   # 群人数增加
EVENT_GROUP_DECREASE = "0022"  # 群人数减少
EVENT_OTHRE = "02DC"  # 禁言、匿名等事件



from .pcapi import (
    send_friend_msg,
    friend_receipt,
    send_group_msg,
    group_receipt,
    upload_friend_image,
    upload_group_image
)

from .webapi import (
    get_group_info,
    set_group_card,
    set_group_shutup,
    get_user_name,
    get_user_cache,
    get_group_cache,
)

from .wtlogin import (
    set_online,
    login_out,
    heatbeat,
    qrcode_login,
    password_login,
    token_login
)



import sys
import json
import base64
import random
import socket
import asyncio
from urllib import parse

import pcqq.utils as utils
import pcqq.const as const
import pcqq.logger as logger
import pcqq.binary as binary


class AsyncTCPClient:
    def __init__(self) -> None:
        domain = random.choice([
            "tcpconn.tencent.com",
            "tcpconn2.tencent.com",
            "tcpconn3.tencent.com",
            "tcpconn4.tencent.com"
        ])
        self.host: str = socket.gethostbyname(domain)
        self.port: int = const.TCP_PORT
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

    async def init(self) -> None:
        self.writer: asyncio.StreamWriter = None
        self.reader: asyncio.StreamReader = None
        self.reader, self.writer = await asyncio.open_connection(
            host=self.host,
            port=self.port,
            loop=self.loop
        )
        logger.info(f"正在创建 TCP 连接: {self.host}:{self.port} -> 成功")

    async def send(self, data: bytes) -> None:
        size = len(data) + 2
        self.writer.write(size.to_bytes(2, "big") + data)

    async def recv(self) -> bytes:
        data = await self.reader.read(const.BUF_SIZE)
        return data[2:]


class QQStruct:
    local_ip: bytes = b''

    server_ip: bytes = b''

    uin: int = 0

    password: bytes = b''

    nickname: str = ''

    is_scancode: bool = False

    login_time: bytes = b''

    redirection_times: int = 0

    redirection_history: bytes = b''

    tgt_key: bytes = b''

    tgtgt_key: bytes = b''

    session_key: bytes = b''

    bkn: int = 0

    skey: str = ''

    cookie: str = ''

    pckey_for_0819: bytes = b''

    pckey_for_0828_send: bytes = b''

    pckey_for_0828_recv: bytes = b''

    token_by_scancode: bytes = b''

    token_0038_from_0825: bytes = b''

    token_0038_from_0818: bytes = b''

    token_0038_from_0836: bytes = b''

    token_0088_from_0836: bytes = b''

    def save_token(self, path: str):
        file = open(path, mode='wb')

        file.write(b'DawnNights'.join([
            self.server_ip,
            self.uin.to_bytes(4, 'big'),
            self.nickname.encode(),
            self.token_0038_from_0836,
            self.token_0088_from_0836,
            self.pckey_for_0828_send,
            self.pckey_for_0828_recv
        ]))

        logger.info(f'已保存登录Token至文件 {path}')
        file.close()

    def load_token(self, path: str):
        file = open(path, mode='rb')
        tokens = file.read().split(b'DawnNights')

        self.pckey_for_0828_recv = tokens.pop()
        self.pckey_for_0828_send = tokens.pop()
        self.token_0088_from_0836 = tokens.pop()
        self.token_0038_from_0836 = tokens.pop()
        self.nickname = tokens.pop().decode()
        self.uin = int.from_bytes(tokens.pop(), 'big')
        self.server_ip = tokens.pop()

        logger.info(f'尝试通过 {path} 进行重连...')
        file.close()


class HttpResponse:
    ok: bool = False
    url: str = ""
    headers: dict = {}
    content: bytes = b""
    status_code: int = 0
    links: parse.ParseResult = None

    def __repr__(self):
        return "<%s Response %d>" % (
            self.links.scheme.upper(),
            self.status_code
        )

    def json(self) -> dict:
        return json.loads(self.content)

    def base64(self) -> str:
        return base64.b64encode(self.content).decode()


class QQClient(QQStruct):
    def __init__(self) -> None:
        self.waiter: dict = {}
        self.plugins: list = []
        self.client: AsyncTCPClient = AsyncTCPClient()
        self.teaer: binary.QQTea = binary.QQTea(bytes())

        self.tgt_key = utils.randbytes(16)
        self.server_ip = socket.inet_aton(self.client.host)

        def run_until_complete(coro):
            task = self.client.loop.create_task(coro)
            self.client.loop.run_until_complete(task)
            return task.result()
        self.run = run_until_complete

        def run_coroutine_threadsafe(coro):
            return asyncio.run_coroutine_threadsafe(
                coro=coro,
                loop=self.client.loop,
            )
        self.run_future = run_coroutine_threadsafe

        self.cache: utils.SqliteDB = utils.SqliteDB("cache.db")
        if self.cache.count("ginfo") == -1:
            self.cache.create(
                "ginfo",
                "name text", "owner integer",
                "group_id integer", "board text",
                "intro text", "mem_num integer"
            )
        if self.cache.count("gmember") == -1:
            self.cache.create(
                "gmember",
                "cord text",
                "user_id integer",
                "name text",
                "group_id integer",
                "op boolen"
            )

    async def write_packet(self, cmd: str, version: str, body: bytes, sequence: bytes = utils.randbytes(2)) -> None:
        writer = binary.Writer()
        writer.write_hex(const.HEADER)

        writer.write_hex(cmd)
        writer.write(sequence)
        writer.write_int32(self.uin)
        writer.write_hex(version)

        if self.session_key:
            writer.write(self.teaer.encrypt(body))
        else:
            writer.write(body)

        writer.write_hex(const.TAIL)
        await self.client.send(writer.clear())

    async def read_packet(self) -> tuple:
        reader = binary.Reader(await self.client.recv())
        reader.read(3)  # packet header

        cmd = reader.read_hex(2)
        sequence = reader.read(2)
        uin = reader.read_int32()  # qq uin
        reader.read(3)  # unknow

        if self.session_key:
            body = self.teaer.decrypt(reader.read()[:-1])
        else:
            body = reader.read()[:-1]

        return cmd, sequence, uin, body

    async def webget(self, url: str, headers: dict = {}, **params) -> HttpResponse:
        ret = parse.urlparse(url)
        if ret.scheme == "http":
            reader, writer = await asyncio.open_connection(
                host=ret.netloc,
                port=80,
            )
        elif ret.scheme == "https":
            reader, writer = await asyncio.open_connection(
                host=ret.netloc,
                port=443,
                ssl=True,
            )

        params = parse.urlencode(params)
        url = url + "?" + params if params else url
        request = f"GET {url} HTTP/1.1\r\n"

        headers["Host"] = ret.netloc
        headers["Connection"] = "close"
        for key, value in headers.items():
            request += f"{key}: {value}\r\n"
        writer.write((request + "\r\n").encode())

        response = HttpResponse()
        response.url = url
        response.links = ret

        # Response Status
        line = await reader.readline()
        response.status_code = int(line.split(b" ")[1])
        response.ok = response.status_code == 200

        # Response Headers
        line = (await reader.readline()).replace(b"\r\n", b"").decode()
        while line:
            key, value = line.split(": ")
            response.headers[key] = value
            line = (await reader.readline()).replace(b"\r\n", b"").decode()

        # Response Data
        response.content = await reader.read()

        writer.close()
        return response

    async def webpost(self, url: str, headers: dict = {}, **params) -> HttpResponse:
        ret = parse.urlparse(url)
        if ret.scheme == "http":
            reader, writer = await asyncio.open_connection(
                host=ret.netloc,
                port=80,
            )
        elif ret.scheme == "https":
            reader, writer = await asyncio.open_connection(
                host=ret.netloc,
                port=443,
                ssl=True,
            )
        else:
            raise Exception("调用方法 webpost 时传入的 url 参数有误")

        request = f"POST {url} HTTP/1.1\r\n"
        if "data" in params and isinstance(params["data"], bytes):
            post_data = params["data"]
        else:
            post_data = parse.urlencode(params).encode()

        headers["Host"] = ret.netloc
        headers["Connection"] = "close"
        headers["Content-Length"] = len(post_data)
        for key, value in headers.items():
            request += f"{key}: {value}\r\n"

        writer.write((request + "\r\n").encode() + post_data)

        response = HttpResponse()
        response.url = url
        response.links = ret

        # Response Status
        line = await reader.readline()
        response.status_code = int(line.split(b" ")[1])
        response.ok = response.status_code == 200

        # Response Headers
        line = (await reader.readline()).replace(b"\r\n", b"").decode()
        while line:
            key, value = line.split(": ")
            response.headers[key] = value
            line = (await reader.readline()).replace(b"\r\n", b"").decode()

        # Response Data
        response.content = await reader.read()

        writer.close()
        return response


sys.modules[__name__] = QQClient()



import time

import pcqq.client as cli
import pcqq.utils as utils
import pcqq.const as const
import pcqq.logger as logger
import pcqq.binary as binary
#cli = cli.QQClient()


async def send_friend_msg(user_id: int, msg_data: bytes, has_image: bool = False):
    """
    发送好友消息

    :param user_id: 好友QQ号

    :param msg_data: PCQQ消息协议数据

    """
    writer = binary.Writer()
    time_stamp = int(time.time()).to_bytes(4, 'big')[::-1]

    writer.write_int32(cli.uin)
    writer.write_int32(user_id)
    writer.write_hex("00 00 00 08 00 01 00")
    writer.write_hex("04 00 00 00 00 36 39")
    writer.write_int32(cli.uin)
    writer.write_int32(user_id)

    writer.write(utils.hashmd5(time_stamp))
    writer.write_hex("00 0B 37 96")
    writer.write(time_stamp)
    writer.write_hex("02 55 00 00 00 00 01 00 00 00")
    writer.write_hex("0C 4D 53 47 00 00 00 00 00")
    writer.write(time_stamp)
    writer.write(time_stamp[::-1])

    writer.write_hex("00 00 00 00 09 00 86 00 00")
    if has_image:
        writer.write_hex("12 E6 B1 89 E4 BB AA E8 9D")
    else:
        writer.write_hex("06 E5 AE 8B E4 BD 93 00 00")
    writer.write(msg_data)

    await cli.write_packet(
        "00 CD",
        const.BODY_VERSION,
        writer.clear()
    )


async def send_group_msg(group_id: int, msg_data: bytes, has_image: bool = False):
    """
    发送群消息

    :param group_id: 目标群号

    :param msg_data: PCQQ消息协议数据

    :param has_image: 消息数据中是否含有图片数据

    """
    writer = binary.Writer()
    time_stamp = int(time.time()).to_bytes(4, 'big')

    if has_image:
        writer.write_hex("00 02 01 00 00 00 00 00 00 00")
    else:
        writer.write_hex("00 01 01 00 00 00 00 00 00 00")

    writer.write_hex("4D 53 47 00 00 00 00 00")
    writer.write(time_stamp)
    writer.write(time_stamp[::-1])
    writer.write_hex("00 00 00 00 09 00 86 00 00")
    writer.write_hex("06 E5 AE 8B E4 BD 93 00 00")
    writer.write(msg_data)
    data = writer.clear()

    writer.__init__()
    writer.write_hex("2A")
    writer.write_int32(utils.gid_from_group(group_id))
    writer.write_int16(len(data))
    writer.write(data)

    await cli.write_packet(
        "00 02",
        const.BODY_VERSION,
        writer.clear()
    )


async def friend_receipt(user_id: int, timestamp: int):
    writer = binary.Writer()
    writer.write_hex("08 01")
    writer.write_hex("12 03 98 01 00")
    writer.write_hex("0A 0E 08")
    writer.write_varint(user_id)
    writer.write_hex("10")
    writer.write_varint(timestamp)
    writer.write_hex("20 00")
    data = writer.clear()

    writer.__init__()
    writer.write_hex("00 00 00 07")
    writer.write_int32(len(data)-7)
    writer.write(data)

    await cli.write_packet(
        "03 19",
        const.FUNC_VERSION,
        writer.clear()
    )


async def group_receipt(group_id: int, msg_id: int):
    writer = binary.Writer()

    writer.write_byte(41)
    writer.write_int32(utils.gid_from_group(group_id))
    writer.write_byte(2)
    writer.write_int32(msg_id)

    await cli.write_packet(
        "00 02",
        const.BODY_VERSION,
        writer.clear()
    )


async def upload_group_image(group_id: int, image):
    writer = binary.Writer()

    writer.write_pbint(group_id, 1)
    writer.write_pbint(cli.uin, 2)
    writer.write_pbint(0, 3)
    writer.write_pbdata(image.hash, 4)
    writer.write_pbint(image.size, 5)
    writer.write_pbhex(
        "37 00 4D 00 32 00 25 00 4C 00 31 00 56 00 32 00 7B 00 39 00 30 00 29 00 52 00", 6
    )
    writer.write_pbint(1, 7)
    writer.write_pbint(1, 9)
    writer.write_pbint(image.width, 10)
    writer.write_pbint(image.height, 11)
    writer.write_pbint(4, 12)
    writer.write_pbdata("26656".encode(), 13)
    data = writer.clear()

    writer.__init__()
    writer.write_pbdata(data, 3)
    data = writer.clear()

    writer.__init__()
    writer.write_pbint(1, 19)
    temp = writer.clear()

    writer.__init__()
    writer.write_pbdata(temp, 2)
    temp = writer.clear()

    writer.__init__()
    writer.write_pbint(1, 1)
    writer.write(temp)
    writer.write_pbint(1, 2)
    writer.write(data)
    data = writer.clear()

    writer.__init__()
    writer.write_hex("00 00 00 07 00 00")
    writer.write_int16(len(data))
    writer.write(data)
    writer.write_hex("70 00 78 03 80 01 00")

    sequence = utils.randbytes(2)
    await cli.write_packet(
        "03 88",
        const.FUNC_VERSION,
        writer.clear(),
        sequence=sequence,
    )

    async def callback(body):
        if len(body) < 128:
            logger.info(f"图片获取成功 -> https://gchat.qpic.cn/gchatpic_new/0/0-0-{image.hash.hex().upper()}/0?term=3")
            image.finish = True
            return  # 无需重新上传

        start = body.find(bytes([66, 128, 1])) + 3
        end = body.find(bytes([128, 128, 8])) - 9
        image.ukey = body[start:end].hex().upper()

        await cli.webpost(
            url="http://htdata2.qq.com/cgi-bin/httpconn?" + "&".join([
                "htcmd=0x6ff0071",
                "ver=5515",
                "term=pc",
                f"ukey={image.ukey}",
                f"filesize={image.size}",
                "range=0",
                f"uin={cli.uin}",
                f"groupcode={group_id}"
            ]),
            data=image.data,
            headers={"User-Agent": "QQClient"}
        )
        logger.info(f"图片上传完成 -> https://gchat.qpic.cn/gchatpic_new/0/0-0-{image.hash.hex().upper()}/0?term=3")
        image.finish = True
    cli.waiter[sequence.hex()] = callback


async def upload_friend_image(user_id: int, image) -> bytes:
    writer = binary.Writer()

    writer.write_pbint(cli.uin, 1)
    writer.write_pbint(user_id, 2)
    writer.write_pbint(0, 3)
    writer.write_pbdata(image.hash, 4)
    writer.write_pbint(image.size, 5)
    writer.write_pbhex(
        "45 00 4F 00 59 00 7B 00 56 00 29 00 45 00 56 00 4A 00 4B 00 48 00 48 00 53 00", 6
    )
    writer.write_pbint(1, 7)
    writer.write_pbint(0, 9)
    writer.write_pbint(image.width, 14)
    writer.write_pbint(image.height, 15)
    data = writer.clear()

    writer.__init__()
    writer.write_pbint(1, 19)
    writer.write_hex("98 03 F6 A8 BF A6 04 A0 03 01")
    temp = writer.clear()

    writer.__init__()
    writer.write_pbdata(data, 2)
    data = writer.clear()

    writer.__init__()
    writer.write_pbdata(temp, 2)
    temp = writer.clear()

    writer.__init__()
    writer.write_pbint(1, 1)
    writer.write(temp)
    writer.write_pbint(1, 1)
    writer.write(data)
    data = writer.clear()

    writer.__init__()
    writer.write_hex("00 00 00 11 00 00")
    writer.write_int16(len(data)-17)
    writer.write(data)

    sequence = utils.randbytes(2)
    await cli.write_packet(
        "03 52",
        const.FUNC_VERSION,
        writer.clear(),
        sequence=sequence,
    )

    async def callback(body):
        reader = binary.Reader(body)

        if reader.tell() < 256:
            reader.read(66)
            image.picid = reader.read(55)  # 无需重新上传
            logger.info(f"图片获取成功 -> https://c2cpicdw.qpic.cn/offpic_new/{cli.uin}/{image.picid.decode()}/0")
            image.finish = True
            return

        reader.read(70)
        image.ukey = reader.read(336).hex().upper()
        reader.read(2)
        image.picid = reader.read(55)

        await cli.webpost(
            url="http://htdata2.qq.com/cgi-bin/httpconn?" + "&".join([
                "htcmd=0x6ff0070",
                "ver=5509",
                f"ukey={image.ukey}",
                f"filesize={image.size}",
                "range=0",
                f"uin={cli.uin}"
            ]),
            data=image.data,
            headers={
                "User-Agent": "QQClient",
            }
        )
        logger.info(f"图片上传完成 -> https://c2cpicdw.qpic.cn/offpic_new/{cli.uin}/{image.picid.decode()}/0")
        image.finish = True

    cli.waiter[sequence.hex()] = callback



import pcqq.const as const
import pcqq.utils as utils
import pcqq.binary as binary


def _tlv(cmd: str, bin: bytes) -> bytes:
    writer = binary.Writer()

    writer.write_hex(cmd)
    writer.write_int16(len(bin))
    writer.write(bin)

    return writer.clear()


def tlv004() -> bytes:
    """
    TLV_0004_QRLogin

    """
    writer = binary.Writer()

    writer.write_hex("00 00")
    writer.write_int16(8)
    writer.write(b'qr_login')

    return _tlv("00 04", writer.clear())


def tlv112(token: bytes) -> bytes:
    """
    TLV_0112_SigIP2

    :param token: Token 0038 From 0825 Protocol Package

    """
    writer = binary.Writer()

    writer.write(token)

    return _tlv("01 12", writer.clear())


def tlv30f() -> bytes:
    """
    TLV_030F_ComputerName

    """
    writer = binary.Writer()

    # ComputerName Length
    writer.write_int16(15)
    # ComputerName Data
    writer.write_hex("44 61 77 6E 4E 69 67 68 74 73 2D 50 43 51 51")

    return _tlv("03 0F", writer.clear())


def tlv005(uin: int) -> bytes:
    """
    TLV_0005_Uin

    :param uin: QQ Number

    """
    writer = binary.Writer()

    writer.write_hex("00 02")   # Tlv Version
    writer.write_int32(uin)

    return _tlv("00 05", writer.clear())


def tlv303(token: bytes) -> bytes:
    """
    TLV_0303_UnknownTag

    :param token: PcToken 0060 From 0819 Protocol Package

    """
    writer = binary.Writer()

    writer.write(token)

    return _tlv("03 03", writer.clear())


def tlv006(
    uin: int,
    tgt_key: bytes,
    md5_once: bytes,
    md5_twice: bytes,
    login_time: bytes,
    local_ip: bytes,
    computer_id: bytes,
    tgtgt: bytes

) -> bytes:
    """
    TLV_0006_TGTGT

    :param uin: QQ Number

    :param tgt_key: TGTKey

    :param md5_once: The password MD5 is encrypted once

    :param md5_twice: The password MD5 is encrypted twice

    :param login_time: Time to Login QQ Server 

    :param local_ip: Local IP Address

    :param computer_id: Computer device ID

    :param tgtgt: TGTGT

    """
    writer = binary.Writer()

    if tgtgt:
        writer.write(tgtgt)
    else:
        writer.write(utils.randbytes(4))
        writer.write_hex("00 02")
        writer.write_int32(uin)
        writer.write_hex(const.SSO_VERSION)  # SSO Version
        writer.write_hex(const.SERVICEID)  # ServiceId
        writer.write_hex(const.CLIENT_VERSION)  # Client Version
        writer.write_hex("00 00")
        writer.write_hex("00")  # Don't remember password
        writer.write(md5_once)
        writer.write(login_time)
        writer.write_hex("00 00 00 00 00 00 00 00 00 00 00 00 00")
        writer.write(local_ip)
        writer.write_hex("00 00 00 00 00 00 00 00 00 10")
        writer.write(computer_id)
        writer.write(tgt_key)

    tea = binary.QQTea(md5_twice)
    return _tlv("00 06", tea.encrypt(writer.clear()))


def tlv015() -> bytes:
    """
    TLV_0015_ComputerGuid

    """
    writer = binary.Writer()

    writer.write_hex("00 01")   # wSubVer

    writer.write_byte(1)
    writer.write_hex("74 83 F2 C3")  # CRC-32 ComputerGuid
    writer.write_int16(16)  # ComputerGuid Length

    # ComputerGuid Data
    writer.write_hex("14 FE 77 FC 00 00 00 00 00 00 00 00 00 00 00 00")

    writer.write_byte(2)
    writer.write_hex("17 65 6E 9D")  # CRC-32 ComputerGuidEX
    writer.write_int16(16)  # ComputerGuidEX Length

    # ComputerGuidEX Data
    writer.write_hex("78 8A 33 DD 00 76 A1 78 EB 8E 5B BB FF 17 D0 10")

    return _tlv("00 15", writer.clear())


def tlv01a(tgtkey: bytes) -> bytes:
    """
    TLV_001A_GTKeyTGTGTCryptedData

    :param tgtkey: Tgtkey

    """

    tea = binary.QQTea(tgtkey)
    return _tlv("00 1A", tea.encrypt(tlv015()))


def tlv018(uin: int, redirect_times: int) -> bytes:
    """
    TLV_0018_Ping

    :param uin: QQ Number

    :param redirect_times: Times of QQ Server redirects

    """
    writer = binary.Writer()

    writer.write_hex("00 01")   # wSubVer
    writer.write_hex(const.SSO_VERSION)  # SSO Version
    writer.write_hex(const.SERVICEID)  # ServiceId
    writer.write_hex(const.CLIENT_VERSION)  # Client Version
    writer.write_int32(uin)
    writer.write_int16(redirect_times)
    writer.write_hex("00 00")

    return _tlv("00 18", writer.clear())


def tlv019() -> bytes:
    """
    TLV_0019_Ping

    """
    writer = binary.Writer()

    writer.write_hex("00 01")   # wSubVer
    writer.write_hex(const.SSO_VERSION)  # SSO Version
    writer.write_hex(const.SERVICEID)  # ServiceId
    writer.write_hex(const.CLIENT_VERSION)  # Client Version
    writer.write_hex("00 00")

    return _tlv("00 19", writer.clear())


def tlv305() -> bytes:
    """
    Tlv_0305_QRCodeParams

    """
    writer = binary.Writer()

    writer.write_hex("00 00 00 00")
    writer.write_hex("00 00 00 05")
    writer.write_hex("00 00 00 04")
    writer.write_hex("00 00 00 00")
    writer.write_hex("00 00 00 48")
    writer.write_hex("00 00 00 02")
    writer.write_hex("00 00 00 02")
    writer.write_hex("00 00")

    return _tlv("03 05", writer.clear())


def tlv114() -> bytes:
    """
    TLV_0114_DHParams

    """
    writer = binary.Writer()

    writer.write_hex(const.EDCH_VERSION)   # Edch Version
    writer.write_int16(25)   # PublicKey Length
    writer.write_hex(const.PUBLICKEY)   # PublicKey Data

    return _tlv("01 14", writer.clear())


def tlv103() -> bytes:
    """
    TLV_0103_SID

    """
    writer = binary.Writer()

    writer.write_hex("00 01 00 10")
    writer.write(utils.randbytes(16))

    return _tlv("01 03", writer.clear())


def tlv312() -> bytes:
    """
    TLV_0312_Misc_Flag

    """
    writer = binary.Writer()

    writer.write_hex("01 00 00 00 00")

    return _tlv("03 12", writer.clear())


def tlv508() -> bytes:
    """
    TLV_0508_UnknownTag

    """
    writer = binary.Writer()

    writer.write_hex("01 00 00 00 02")

    return _tlv("05 08", writer.clear())


def tlv313() -> bytes:
    """
    TLV_0313_GUID_Ex

    """
    writer = binary.Writer()

    writer.write_hex("01")
    writer.write_byte(1)    # GUID_Ex Count

    writer.write_hex("02")  # GUID_Ex Index
    writer.write_int16(16)  # GUID_Ex Length
    # GUID_Ex Data
    writer.write_hex("EE 47 7F A4 BC D6 EE 65 02 65 4D E9 43 38 4C 3D")

    writer.write_hex("00 00 00 EB")  # System Tick

    return _tlv("03 13", writer.clear())


def tlv102(token: bytes) -> bytes:
    """
    TLV_0102_Official

    : param token: Token 0038 From 0825 Protocol Package

    """
    writer = binary.Writer()

    writer.write_hex("00 01")
    writer.write(utils.randbytes(16))
    writer.write_int16(len(token))
    writer.write(token)
    writer.write_hex("00 14")
    writer.write(utils.randbytes(20))

    return _tlv("01 02", writer.clear())


def tlv309(
    server_ip: bytes,
    redirection_history: bytes,
    redirection_times: int
) -> bytes:
    """
    TLV_0309_Ping_Strategy

    :param server_ip: QQ Server IP Address

    :param redirect_history: History of QQ Server redirects

    :param redirect_times: Times of QQ Server redirects

    """
    writer = binary.Writer()

    writer.write_hex("00 01")
    writer.write(server_ip)

    if redirection_times:
        writer.write_byte(redirection_times)
        writer.write(redirection_history)
    else:
        writer.write_hex("00 02")

    return _tlv("03 09", writer.clear())


def tlv036() -> bytes:
    """
    TLV_0036_LoginReason

    """
    writer = binary.Writer()

    writer.write_hex("00 02")
    writer.write_hex("00 01")
    writer.write_hex("00 00 00 00 00 00 00 00 00 00 00 00 00 00")

    return _tlv("00 36", writer.clear())


def tlv007(token: bytes) -> bytes:
    """
    TLV_0007_TGT

    :param token: PcToken 0088 From 0836 Protocol Package

    """
    writer = binary.Writer()

    writer.write(token)

    return _tlv("00 07", writer.clear())


def tlv00c(server_ip: bytes) -> bytes:
    """
    TLV_000C_PingRedirect

    :param server_ip: QQ Server IP Address

    """
    writer = binary.Writer()

    writer.write_int16(2)
    writer.write_hex("00 01 00 00 00 00 00 00 00 00")
    writer.write(server_ip)
    writer.write_int16(80)
    writer.write_hex("00 00 00 00")

    return _tlv("00 0C", writer.clear())


def tlv01f() -> bytes:
    """
    TLV_001F_DeviceID

    """
    writer = binary.Writer()

    writer.write_hex("00 01")
    writer.write(utils.randbytes(32))

    return _tlv("00 1F", writer.clear())


def tlv105() -> bytes:
    """
    TLV_0105_m_vec0x12c

    """
    writer = binary.Writer()

    writer.write_int16(1)
    writer.write_hex("01 02")

    writer.write_int16(20)
    writer.write_hex("01 01")
    writer.write_int16(16)
    writer.write(utils.randbytes(16))

    writer.write_int16(20)
    writer.write_hex("01 02")
    writer.write_int16(16)
    writer.write(utils.randbytes(16))

    return _tlv("01 05", writer.clear())


def tlv10b() -> bytes:
    """
    TLV_010B_QDLoginFlag

    """
    writer = binary.Writer()

    writer.write_hex("00 02")
    writer.write(utils.randbytes(17))   # (Random)ClientMd5 + QDFlag
    writer.write_hex("10 00 00 00 00 00 00 00 02")

    writer.write_hex("00 63")
    writer.write_hex("3E 00 63 02")
    writer.write_hex(const.DWQD_VERSION)
    writer.write_hex("00 04 00")
    writer.write(utils.randbytes(2))
    writer.write_hex("00 00 00 00")
    writer.write(utils.randbytes(16))

    writer.write_hex("01 00")
    writer.write_hex("00 00 00 00")
    writer.write_hex("00 00")
    writer.write_hex("00 00 00 00 00 00 00 00")

    writer.write_hex("00 00 00 01")
    writer.write_hex("00 00 00 01")
    writer.write_hex("00 00 00 01")
    writer.write_hex("00 00 00 01")
    writer.write_hex("00 FE 26 81 75 EC 2A 34 EF")

    # QDDATA
    writer.write_hex("02 3E 50 39 6D B1 AF CC 9F EA 54 E1")
    writer.write_hex("70 CC 6C 9E 4E 63 8B 51 EC 7C 84 5C")

    writer.write_hex("68")
    writer.write_hex("00 00 00 00")

    return _tlv("01 0B", writer.clear())


def tlv02d(local_ip: bytes) -> bytes:
    """
    TLV_002D_LocalIP

    """
    writer = binary.Writer()

    writer.write_hex("00 01")
    writer.write(local_ip)

    return _tlv("00 2D", writer.clear())



from .qqtea import QQTea
from .reader import Reader
from .writer import Writer
from .qqtlv import (
    tlv112,
    tlv30f,
    tlv004,
    tlv005,
    tlv303,
    tlv006,
    tlv015,
    tlv01a,
    tlv018,
    tlv019,
    tlv305,
    tlv114,
    tlv103,
    tlv312,
    tlv508,
    tlv313,
    tlv102,
    tlv309,
    tlv036,
    tlv007,
    tlv00c,
    tlv01f,
    tlv105,
    tlv10b,
    tlv02d
)


from setuptools import setup, find_packages



packages = find_packages()

setup(
    name='py-pcqq',
    version='1.3.0',
    url='https://github.com/DawnNights/py-pcqq',
    author='DawnNights',
    author_email='2224825532@qq.com',
    packages=packages,
    description=u'一个使用pcqq协议的简易python qqbot库',
    long_description='''

简介
----

PY-PCQQ 是一个基于 **QQ轻聊版 7.9** 客户端协议的 Python 异步 QQ
机器人支持库，它会对 QQ
服务端发出的协议包进行解析和处理，并以插件化的形式，分发给消息所对应的命令处理器。

除了起到解析消息的作用，PY-PCQQ 还通过
装饰器、异步、回调等方式实现了一套简洁易用的会话机制和插件机制，以便于用户快速上手。

PY-PCQQ 在其底层与 QQ 服务端实现交互的部分使用的是标准库
``asyncio.open_connection`` 所创建的异步 TCP
连接，这意味着在本协议库提供的内容多为异步操作，在调用相关函数或方法时应注意加上
**await** 关键字。

本项目基本仅由 Python3
的标准库所实现，但若是在没有图形界面的系统中使用扫码登录，需要自行安装第三方库
``pillow`` 使得程序能在终端环境中打印登录二维码。值得一提的是，在
Android 手机等移动设备中，你也可以通过
`pydroid3 <https://apkdownloadforandroid.com/ru.iiec.pydroid3/>`__
这样的应用来安装运行本协议库，例如:
`在手机上玩转QQ机器人？ <https://b23.tv/ZVHP0lK>`__

最后要说的是，这个项目仅仅只是本废物空闲之余的兴趣之作，存在着大量不成熟与不完善的地方。如果对
QQ 机器人开发有所需求，可以移步至更加强大与稳定的
`mirai <https://github.com/mamoe/mirai>`__ 或
`go-cqhttp <https://github.com/Mrs4s/go-cqhttp/>`__ 等项目。

已实现功能
----------

登录
^^^^

-  ☒ 账号密码登录
-  ☒ 二维码登录
-  ☒ 本地Token重连

发送消息
^^^^^^^^

-  ☒ At
-  ☒ 文本
-  ☒ 表情
-  ☒ xml卡片
-  ☒ 图片

接收消息
^^^^^^^^

-  ☒ At
-  ☒ 文本
-  ☒ 图片
-  ☒ 表情

接收事件
^^^^^^^^

-  ☒ 群消息
-  ☒ 好友消息
-  ☒ 进群事件
-  ☒ 退群事件
-  ☒ 禁言事件

其它操作
^^^^^^^^

-  ☒ 修改群成员Card
-  ☒ 设置群成员禁言

文档
----

暂时咕咕中，请自行查看项目中的 ``example.py`` 查看案例

'''
    
)



import sqlite3

class SqliteDB:
    def __init__(self, path: str) -> None:
        db = sqlite3.connect(path)
        cu = db.cursor()

        self.close = db.close
        self.commit = db.commit
        self.execute = cu.execute
        self.fetchone = cu.fetchone
        self.fetchall = cu.fetchall

    def create(self, table_name: str, *args) -> bool:
        try:
            self.execute("create table %s %s" % (
                table_name, str(args).replace("'", "")
            ))
            self.commit()
            return True
        except:
            return False

    def insert(self, table_name: str, *args):
        '''调用本方法后需手动调用commit方法'''
        args = [f"'{arg}'" if isinstance(arg, str) else str(arg) for arg in args]
        self.execute("insert into %s values (%s)" % (
            table_name, ", ".join(args)
        ))

    def select(self, table_name: str, **params) -> list:
        self.execute("select * from %s where %s" % (
            table_name, " and ".join(
                [f"{key}={params[key]}" for key in params])
        ))
        return self.fetchall()

    def exist(self, table_name: str, **params) -> bool:
        return bool(self.select(table_name, **params))

    def update(self, table_name: str, *args, **params):
        args = " ".join(args)
        params = " ".join([f"{key}={params[key]}" for key in params])
        self.execute("update %s set %s where %s" % (
            table_name, args, params
        ))
        self.commit()

    def delete(self, table_name: str, **params):
        self.execute("select * from %s where %s" % (
            table_name, " ".join([f"{key}={params[key]}" for key in params])
        ))
        self.commit()

    def count(self, table_name: str) -> int:
        try:
            self.execute("select count(*) from %s" % (table_name))
            return self.fetchone()[0]
        except sqlite3.OperationalError:
            return -1


import pcqq.client as cli
import pcqq.network as net
import pcqq.message as msg
import pcqq.logger as logger
# cli = cli.QQClient()


class Session:
    msg_id: int = 0
    timestamp: int = 0

    self_id: int = 0
    user_id: int = 0
    group_id: int = 0
    target_id: int = 0

    matched = None
    message: str = ""
    event_type: str = ""
    raw_message: list = []

    async def send_msg(self, *message):
        if self.group_id:
            await self.send_group_msg(self.group_id, *message)
        else:
            await self.send_friend_msg(self.user_id, *message)

    async def send_group_msg(self, group_id, *message):
        data = b""
        escape = ""
        has_image = False

        for seg in message:
            if isinstance(seg, str):
                escape += seg
                data += msg.text(seg)

            if isinstance(seg, dict) and "type" in seg:
                if seg["type"] == "at" and "qq" in seg:
                    data += await msg.at(int(seg["qq"]), group_id)
                    escape += f"[PQ:at,qq={seg['qq']}]"

                elif seg["type"] == "face" and "id" in seg:
                    data += msg.face(int(seg["id"]))
                    escape += f"[PQ:face,qq={seg['id']}]"
                elif seg["type"] == "xml" and "code" in seg:
                    data += msg.xml(seg["code"])
                    escape += f"[PQ:xml,code={seg['code']}]"
                elif seg["type"] == "music":
                    if "keyword" in seg:
                        data += await msg.qqmusic(str(seg["keyword"]))
                        escape += f"[PQ:music,keyword={seg['keyword']}]"
                    else:
                        del seg["type"]
                        data += msg.music(**seg)
                        escape += f"[PQ:music,{','.join([k+'='+v for k,v in seg.items()])}]"
                elif seg["type"] == "image":
                    if "data" in seg and isinstance(seg["data"], bytes):
                        image = msg.Image(seg["data"])

                    elif "url" in seg and seg["url"].startswith("http"):
                        image = msg.Image((await cli.webget(seg["url"])).content)
                    else:
                        continue

                    data += await msg.image_group(group_id, image)
                    escape += f"[PQ:image,file={image.hash.hex().upper()}]"
                    has_image = True

        await net.send_group_msg(group_id, data, has_image)
        logger.info(f"发送群消息: %s -> %s(%d)" % (
            escape, await net.get_group_cache(group_id), group_id
        ))

    async def send_friend_msg(self, user_id, *message):
        data = b""
        escape = ""
        has_image = False
        for seg in message:
            if isinstance(seg, str):
                escape += seg
                data += msg.text(seg)

            if isinstance(seg, dict) and "type" in seg:
                if seg["type"] == "face" and "id" in seg:
                    data += msg.face(int(seg["id"]))
                    escape += f"[PQ:face,qq={seg['id']}]"
                elif seg["type"] == "xml" and "code" in seg:
                    data += msg.xml(seg["code"])
                    escape += f"[PQ:xml,code={seg['code']}]"
                elif seg["type"] == "music":
                    if "keyword" in seg:
                        data += await msg.qqmusic(str(seg["keyword"]))
                        escape += f"[PQ:music,keyword={seg['keyword']}]"
                    else:
                        del seg["type"]
                        data += msg.music(**seg)
                        escape += f"[PQ:music,{','.join([k+'='+v for k,v in seg.items()])}]"
                elif seg["type"] == "image":
                    if "data" in seg and isinstance(seg["data"], bytes):
                        image = msg.Image(seg["data"])

                    elif "url" in seg and seg["url"].startswith("http"):
                        image = msg.Image((await cli.webget(seg["url"])).content)
                    else:
                        continue

                    data += await msg.image_friend(user_id, image)
                    escape += f"[PQ:image,file={image.hash.hex().upper()}]"
                    has_image = True

        await net.send_friend_msg(user_id, data, has_image)
        logger.info(f"发送好友消息: %s -> %s(%d)" % (
            escape, await net.get_user_cache(user_id), user_id
        ))



import io


def varint(num: int) -> bytes:
    val = bytearray()
    while num > 127:
        val.append(0x80 | num & 0x7F)
        num = num >> 7
    val.append(num)
    return bytes(val)


class Writer:
    def __init__(self) -> None:
        self.raw = io.BytesIO()

    def clear(self) -> bytes:
        data = self.raw.getvalue()
        self.raw.close()
        return data

    def write(self, bin: bytes) -> None:
        self.raw.write(bin)
    
    def write_int16(self, num: int) -> None:
        self.write(num.to_bytes(2, 'big'))

    def write_int32(self, num: int) -> None:
        self.write(num.to_bytes(4, 'big'))

    def write_hex(self, hex: str) -> None:
        self.write(bytes.fromhex(hex))

    def write_byte(self, byte: int) -> None:
        self.write(byte.to_bytes(1, 'big'))

    def write_varint(self, num: int) -> None:
        self.write(varint(num))

    def write_pbint(self, num: int, mark: int, isid: bool = True):
        mark = 8 * mark if isid else mark
        self.write_varint(mark)
        self.write_varint(num)

    def write_pbdata(self, data: bytes, mark: int, isid: bool = True):
        mark = 2 + 8 * mark if isid else mark
        self.write_varint(mark)
        self.write_byte(len(data))
        self.write(data)

    def write_pbhex(self, hex: str, mark: int, isid: bool = True):
        return self.write_pbdata(bytes.fromhex(hex), mark, isid)
